package diego


import diego.init.Settings
import diego.tooling.IndexRouteHandler
import diego.utils.Timer
import diego.web.HttpRequestHandler
import diego.web.impl.RouterImpl
import groovy.util.logging.Slf4j
import io.vertx.core.http.HttpServer
import io.vertx.ext.web.Router
import io.vertx.ext.web.handler.StaticHandler

import static io.vertx.core.http.HttpMethod.PATCH
import static io.vertx.core.http.HttpMethod.POST
import static io.vertx.core.http.HttpMethod.PUT


/**
 * Bootstraps the application and initializes the injector.
 *
 * Author: Kevin Sheppard
 * Date: February 18, 2023
 */
@Slf4j
class Diego {
    private static Settings settings
    private static FrameworkModule frameworkModule
    private static Router router
    private static HttpServer server
    private static Timer startupTimer = new Timer()

    /**
     * Extracted this out so that Diego.run and CLILauncher.main can perform different
     * bootstrapping processes.
     *
     * @param runnable Code that runs once the server started successfully.
     */
    protected static void start(Runnable runnable) {
        server.listen {result ->
            if(result.succeeded()) {
                log.info("Listening on http://localhost:${settings.http.port} (Bound to all interfaces on 0.0.0.0)")
                log.trace("Application started in ${startupTimer.stopClock().elapsedTime}s")
                if(runnable) runnable.run()
            } else {
                log.error(result.cause().message)
                throw result.cause()
            }
        }
    }

    /**
     * Bootstraps the core framework components and starts the application server.
     *
     * @param appClass A class that implements {@link diego.Application}
     */
    static void run(Class<Application> appClass) {
        bootstrap()
        final Application application = Components.createInstance(appClass)
        configureRouter(application)
        start(application::onStart)
    }

    /**
     * Initializes all of the core components of the framework, including the dependency container.
     */
    protected static void bootstrap() {
        startupTimer.startClock()
        settings = Settings.instance
        frameworkModule = FrameworkModule.create(settings)
        Components.initializeWith(frameworkModule)
        /* We create an instance of router this way so that interceptors for the HttpRequestHandler will work. */
        router = Components.createInstance(HttpRequestHandler).delegate
        server = frameworkModule.server
        server.requestHandler(router)
        Components.initializeWith(frameworkModule)
    }

    protected static void configureRouter(Application application) {
        /* Serve WebJars */
        router.get('/webjars/*').handler(StaticHandler.create('META-INF/resources/webjars').setCachingEnabled(settings.env.prod))
        if(settings.env.dev) {
            /* Add a route that shows the list of routes */
            if (settings.env.dev) router.get("/_routes").handler(new IndexRouteHandler(router))
        }
        [POST, PUT, PATCH].forEach(method -> {
            router.route().method(method)
                .handler(frameworkModule.bodyHandler())
        })
        application.configure(new RouterImpl(router))
    }

    protected static Router getRouter() {
        return router
    }
}
