package diego

import com.fasterxml.jackson.databind.DeserializationFeature
import com.fasterxml.jackson.databind.ObjectMapper
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule
import com.google.inject.AbstractModule
import com.google.inject.Provides
import com.google.inject.Singleton
import diego.db.DatabaseModule
import diego.db.SessionFactoryProvider
import diego.init.Settings
import diego.templates.FreemarkerModule
import groovy.util.logging.Slf4j
import io.vertx.core.Vertx
import io.vertx.core.VertxOptions
import io.vertx.core.http.HttpServer
import io.vertx.core.http.HttpServerOptions
import io.vertx.ext.web.Router
import io.vertx.ext.web.handler.BodyHandler
import jakarta.validation.Validation
import jakarta.validation.Validator
import jakarta.validation.ValidatorFactory
import org.hibernate.SessionFactory

import java.text.SimpleDateFormat
/**
 * The Guice Module that holds all of the core components of the framework.
 *
 * @author Kevin Sheppard
 * Date: Mar 14, 2022
 */
@Slf4j
class FrameworkModule extends AbstractModule {
  private SessionFactory sessionFactory
  private ValidatorFactory validatorFactory
  private Vertx vertx
  private final Settings settings

  private FrameworkModule(Settings settings) {
    this.settings = settings
  }

  /**
   * Creates in instance of this class with all of its heavy dependencies already initialized.
   *
   * It's done this way so that they will not have to be re-initialized
   * should we need to clear the injector after we perform a hot reload.
   *
   * @return A new instance of this class, but with all if its heavy dependencies initialized
   */
  static FrameworkModule create(Settings settings) {
    final FrameworkModule module = new FrameworkModule(settings)
    module.vertx = Vertx.vertx(new VertxOptions(workerPoolSize: 2))
    module.validatorFactory = Validation.buildDefaultValidatorFactory()
    if(settings.data.hasPath("app.db"))
      module.sessionFactory = new SessionFactoryProvider(settings).get()
    return module
  }

  @Override
  protected void configure() {
    super.configure()
    bind(Settings).toInstance(settings)
    bind(Vertx).toInstance(vertx)
    if(sessionFactory) {
      install(new DatabaseModule(sessionFactory))
    }
    install(new FreemarkerModule())
    settings.modules.each(this::install)
  }

  Vertx getVertx() {
    return vertx
  }

  @Provides
  Validator validator() {
    return validatorFactory.validator
  }

  @Provides @Singleton
  ObjectMapper objectMapper() {
    final ObjectMapper mapper = new ObjectMapper()
    mapper
        .disable(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES)
        .setDateFormat(new SimpleDateFormat(settings.dateFormat))
        .registerModule(new JavaTimeModule())
    return mapper
  }

  @Provides
  Router getRouter() {
    return Router.router(this.vertx)
  }

  /**
   *
   * @return An HttpServer configured according to application settings.
   */
  HttpServer getServer() {
    final var serverOptions = new HttpServerOptions(port: settings.http.port, host: '0.0.0.0')
    return this.vertx.createHttpServer(serverOptions)
  }

  @Provides @Singleton
  BodyHandler bodyHandler() {
    BodyHandler.create(true)
        .setMergeFormAttributes(false)
        .setBodyLimit(settings.http.maxBodySize)
        .setUploadsDirectory(settings.http.fileUploadsDir)
        .setDeleteUploadedFilesOnEnd(settings.http.autoDeleteFiles)
  }
}
