package diego.db

import com.google.common.reflect.ClassPath
import com.google.inject.Inject
import com.google.inject.Provider
import com.typesafe.config.Config
import diego.init.Settings
import groovy.util.logging.Slf4j
import org.hibernate.SessionFactory
import org.hibernate.boot.MetadataSources
import org.hibernate.boot.model.naming.CamelCaseToUnderscoresNamingStrategy
import org.hibernate.boot.registry.StandardServiceRegistry
import org.hibernate.boot.registry.StandardServiceRegistryBuilder
import org.hibernate.cfg.Environment

import javax.persistence.Entity

/**
 * A provider for a Hibernate SessionFactory.
 *
 * @author Kevin Sheppard
 * Date: Feb 22, 2022
 */
@Slf4j
class SessionFactoryProvider implements Provider<SessionFactory> {

  private final Settings settings
  private final StandardServiceRegistryBuilder serviceRegistryBuilder = new StandardServiceRegistryBuilder()
  private final Set<String> packagesToScan = new HashSet<>()

  @Inject
  SessionFactoryProvider(Settings settings) {
    this.settings = settings
  }

  /**
   * A constructor used mainly for testing.
   *
   * @param user The database username
   * @param password The database password
   * @param url The JDBC URL
   * @param driver The database driver class
   */
  SessionFactoryProvider(String user, String password, String url, Class<?> driver) {
    this.settings = null
    serviceRegistryBuilder.tap {
      applySetting(Environment.USER, user)
      applySetting(Environment.PASS, password)
      applySetting(Environment.URL, url)
      applySetting(Environment.DRIVER, driver.canonicalName)
      applySetting(Environment.HBM2DDL_AUTO, "update")
      applySetting(Environment.PHYSICAL_NAMING_STRATEGY, CamelCaseToUnderscoresNamingStrategy.canonicalName)
      applySetting(Environment.SHOW_SQL, true)
      applySetting(Environment.FORMAT_SQL, true)
      applySetting(Environment.CURRENT_SESSION_CONTEXT_CLASS, "thread")
    }
  }

  @Override
  SessionFactory get() {
    return this.settings ? buildSessionFactoryFromSettings() : buildSessionFactory()
  }

  SessionFactoryProvider addPackageToScan(String packageName) {
    packagesToScan.add(packageName)
    return this
  }

  private SessionFactory buildSessionFactory() {
    final StandardServiceRegistry standardServiceRegistry = serviceRegistryBuilder.build()
    final var metadataSources = new MetadataSources(standardServiceRegistry)
    packagesToScan.forEach(pkg -> registerEntities(metadataSources, pkg))
    return metadataSources.buildMetadata().buildSessionFactory()
  }

  private static void registerEntities(MetadataSources sources, String packageName) {
    ClassPath.from(SessionFactoryProvider.getClassLoader())
            .getTopLevelClasses(packageName)
            .stream()
            .map(classInfo -> classInfo.load())
            .filter(klass -> klass.getAnnotation(Entity) != null )
            .peek(klass -> log.trace('Found JPA entity {}', klass.name))
            .forEach(sources::addAnnotatedClass)
  }

  private SessionFactory buildSessionFactoryFromSettings() {
    final var env = settings.env
    Config dbConfig = settings.data.getConfig("app.db")
    final StandardServiceRegistry standardServiceRegistry = serviceRegistryBuilder.tap {
      applySetting(Environment.DRIVER, dbConfig.getString("driver"))
      applySetting(Environment.URL, dbConfig.getString("url"))
      applySetting(Environment.USER, dbConfig.getString("user"))
      applySetting(Environment.PASS, dbConfig.getString("password"))
      if(dbConfig.hasPath("dialect"))
        applySetting(Environment.DIALECT, dbConfig.getString("dialect"))
      applySetting(Environment.HBM2DDL_AUTO, env.isProd() ? "validate" : "update")
      applySetting(Environment.PHYSICAL_NAMING_STRATEGY, CamelCaseToUnderscoresNamingStrategy.canonicalName)
      applySetting(Environment.SHOW_SQL, !env.isProd())
      applySetting(Environment.FORMAT_SQL, !env.isProd())
      applySetting(Environment.CURRENT_SESSION_CONTEXT_CLASS, "thread")
      // Apply any extra JPA/Hibernate settings not covered here.
      if(dbConfig.hasPath("ext")) {
        final Config extraSettings = dbConfig.getConfig("ext")
        extraSettings.entrySet().each { entry -> applySetting(entry.key, entry.value.unwrapped())}
      }
      return it
    }.build()
    final var metadataSources = new MetadataSources(standardServiceRegistry)
    dbConfig.getStringList("entities").forEach(metadataSources::addAnnotatedClassName)
    if(dbConfig.hasPath("packagesToScan")) {
      dbConfig.getStringList("packagesToScan").forEach(pkg -> registerEntities(metadataSources, pkg))
    }
    return metadataSources.buildMetadata().buildSessionFactory()
  }

}
