package diego.db

import com.google.inject.Inject
import com.google.inject.Singleton
import groovy.util.logging.Slf4j
import org.hibernate.HibernateException
import org.hibernate.Session
import org.hibernate.SessionFactory

import javax.persistence.TransactionRequiredException

/**
 * Created so that we can have more granular control over when the database session closes.
 *
 * Using SessionFactory.currentSession() is convenient, but it closes the session after every commit.
 * This makes it impossible to have lazy loaded associations as the session needs to be open to fetch associated
 * rows from the database.
 *
 * We also need to have control over when transactions actually begin and end.
 *
 * We have to roll up our sleeves and commit to building a true session-per-request model.
 *
 * @author Kevin Sheppard
 * Date: Aug 17, 2022
 */
@Singleton
@Slf4j
class SessionManager implements Closeable {
  private SessionFactory sessionFactory
  private ThreadLocal<Session> sessionThreadLocal = new ThreadLocal<>()

  @Inject
  SessionManager(SessionFactory sessionFactory) {
    this.sessionFactory = sessionFactory
  }

  /**
   *
   * @return A new Session or the existing session if one is open. Thread safe 👍🏽
   */
  Session getSession() {
    var session = sessionThreadLocal.get()
    if(session) return session
    session = sessionFactory.openSession()
    sessionThreadLocal.set(session)
    return session
  }

  /**
   * Start a transaction.
   */
  void beginTransaction() {
    if(session.transaction.active) {
      log.debug("Transaction already in progress.")
      return
    }
    session.beginTransaction()
  }

  /**
   * Commit any existing transactions.
   */
  void commit() {
    final var session = sessionThreadLocal.get()
    if(!session) return
    final var tx = session.transaction
    try {
      if(tx.active) {
        log.trace("Committing transaction {}", tx)
        tx.commit()
      }
    } finally {
      if(tx.active) tx.rollback()
    }
  }

  /**
   * Close the session completely.
   */
  void close() {
    try {
      final var session = sessionThreadLocal.get()
      if(session) session.close()
    } catch (HibernateException e) {
      log.error(e.message, e)
    } finally {
      sessionThreadLocal.remove()
    }
  }
}
