package diego.db;

import java.io.Serializable;
import java.util.List;
import java.util.Map;

/**
 * An abstraction for 99% of the database operations I've done over the years.
 *
 * @author Kevin Sheppard
 * Date: July 07, 2023
 */
public interface SqlQuery {
    /**
     * Set the parameters of the query from the properties of the Bean.
     *
     * @param bean A regular Java Bean
     * @return This instance
     */
    SqlQuery bind(Object bean);

    /**
     * Execute an update query (insert, update, delete).
     */
    void executeUpdate();

    /**
     *
     * @return A Map containing the data from the query.
     */
    Map<String, Object> findRow();

    /**
     * Execute a query and map it to a Java Bean. The bean must have a default no-args constructor.
     *
     * @param type The class of the Java bean
     * @return An instance of T
     * @param <T> The type of bean to return
     */
    <T> T findRow(Class<T> type);

    /**
     *
     * @return A List of Maps containing the data from the query.
     */
    List<Map<String, Object>> findRows();

    /**
     * Execute a query and return a list of Beans of type T
     *
     * @param type The class of the Java bean
     * @return An instance of T
     * @param <T> The type of bean to return
     */
    <T> List<T> findRows(Class<T> type);

    /**
     * Execute a query that returns a single scalar value, such as a VARCHAR column or the result of a count()
     *
     * @param type The scalar type
     * @return A scalar value of the specified type
     * @param <T> A Serializable class such as String, Integer or Long
     */
    <T extends Serializable> T findScalar(Class<T> type);

    /**
     * Execute a query that returns a List of scalar values, such as a list of VARCHAR rows
     *
     * @param type The scalar type
     * @return A list of scalar values of the specified type
     * @param <T> A Serializable class such as String, Integer or Long
     */
    <T extends Serializable> List<T> findScalarList(Class<T> type);

    /**
     * Set the parameters of this query from a Map
     *
     * @param namedParams a Map of parameters
     * @return This instance
     */
    SqlQuery params(Map<String, Object> namedParams);

    /**
     * Set the parameters of this query using positional arguments
     *
     * @param positionalParams A list of parameters
     * @return This instance
     */
    SqlQuery params(Object... positionalParams);

    SqlQuery setMaxRows(int maxRows);

    SqlQuery setFirstRow(int firstRow);

}
