package diego.db

import com.fasterxml.jackson.databind.ObjectMapper
import diego.utils.PropertyMapper
import groovy.util.logging.Slf4j
import org.hibernate.Session
import org.hibernate.query.Query
import org.hibernate.query.QueryParameter

import javax.persistence.NoResultException
import javax.persistence.Tuple
import javax.transaction.Transactional
import java.beans.PropertyDescriptor
import java.util.stream.Collectors

@Slf4j
class SqlQueryImpl implements SqlQuery {
    private final Session session
    private final String sql
    private Query<Tuple> query

    SqlQueryImpl(Session session, String sql) {
        this.session = session
        this.sql = sql
    }

    SqlQuery bind(Object bean) {
        final var _query = getQuery()
        // Get all the query parameters in a List for easy searching
        List<String> queryParameters = _query.parameterMetadata.namedParameters.stream().map(QueryParameter::getName).collect(Collectors.toList())
        // Filter the property descriptors to only those needed to set parameters on the query
        List<PropertyDescriptor> propertyDescriptors = PropertyMapper.getPropertyDescriptors(bean)
                .stream()
                .filter(pd -> queryParameters.contains(pd.name)).collect(Collectors.toList())
        // Ensure that there are an equal amount of property descriptors for the amount of query parameters
        assert queryParameters.size() == propertyDescriptors.size() : 'Unequal amount of parameters and bean properties'
        for(PropertyDescriptor pd : propertyDescriptors) {
            assert pd.readMethod != null : "No getter for property ${pd.name}"
            _query.setParameter(pd.name, pd.readMethod.invoke(bean))
        }
        return this
    }

    @Transactional
    void executeUpdate() {
        getQuery().executeUpdate()
    }

    /**
     * @return A single row from the database that matches the criteria of the query.
     */
    Map<String, Object> findRow() {
        try {
            return tupleToMap(getQuery().singleResult)
        } catch (NoResultException ignored) {
            return new HashMap<String, Object>()
        }
    }

    <T> T findRow(Class<T> type) {
        try {
            Map<String, Object> row = findRow()
            if(row.isEmpty()) return null
            return mapToBean(type, row)
        } catch (Exception e) {
            log.error(e.message, e)
            return null
        }
    }

    /**
     * @return Rows from the database matching the criteria of the query.
     */
    List<Map<String, Object>> findRows() {
        return getQuery().resultStream.map(this::tupleToMap).collect(Collectors.toList())
    }

    <T> List<T> findRows(Class<T> type) {
        var result =  findRows()
        result.stream().map(row -> mapToBean(type, row)).collect(Collectors.toList())
    }

    <T extends Serializable> T findScalar(Class<T> resultType) {
        try {
            return getQuery().singleResult.get(0, resultType)
        } catch (NoResultException ignored) {
            return null;
        }
    }

    <T extends Serializable> List<T> findScalarList(Class<T> type) {
        getQuery().resultStream.map(tuple -> tuple.get(0, type)).collect(Collectors.toList())
    }

    /**
     * Creates an instance of T filled with the data from a database query.
     *
     * @param type The Bean type
     * @param row Data retrieved from the database
     * @return An instance of T with its values set from the row data
     */
    protected static  <T> T mapToBean(Class<T> type, Map<String, Object> row) throws IllegalArgumentException {
        return new ObjectMapper().convertValue(row, type)
    }

    private Query getQuery() {
        if(!this.query) this.query = session.createNativeQuery(sql, Tuple)
        return this.query
    }

    <T> List<T> findValueList(Class<T> type) {
        return getQuery().resultList as List<T>
    }

    SqlQueryImpl param(String name, Object value) {
        getQuery().setParameter(name, value)
        return this
    }

    SqlQueryImpl params(Object... positionalParams) {
        final var _query = getQuery()
        for(int i = 0; i<positionalParams.length; i++) {
            _query.setParameter((i+1), positionalParams[i])
        }
        return this
    }

    SqlQueryImpl params(Map<String, Object> namedParameters) {
        final var _query = getQuery()
        for(Map.Entry<String, Object> entry : namedParameters.entrySet()) {
            _query.setParameter(entry.key, entry.value)
        }
        return this
    }

    SqlQuery setMaxRows(int maxRows) {
        getQuery().setMaxResults(maxRows)
        return this
    }

    SqlQuery setFirstRow(int firstRow) {
        getQuery().setFirstResult(firstRow)
        return this
    }
/**
     * <p>Converts a snake_case String to camelCase. If there is a camel case pattern anywhere in the input string, the input
     * string is returned as is.</p>
     *
     * <p>At the very least, the input string is converted to lower case.</p>
     *
     * @param str The input string.
     * @return The input string in camelCase.
     */
    static String toCamelCase(String str) {
        if(str.find("[a-z][A-Z]")) return str // If it's already camel case, return it
        var lowerCaseString = str.toLowerCase()
        while(lowerCaseString.contains("_")) {
            var index = lowerCaseString.indexOf("_") + 1
            var upperCaseChar = new String(Character.toUpperCase(lowerCaseString.charAt(index)))
            lowerCaseString = lowerCaseString.replaceFirst("_[a-z]", upperCaseChar)
        }
        return lowerCaseString
    }

    /**
     * Converts a Tuple to a Map.
     *
     * @param tuple A Tuple.
     * @return A Map of the tuple's aliases and values.
     */
    private static Map<String, Object> tupleToMap(Tuple tuple) {
        Map<String, Object> result = new LinkedHashMap()
        tuple.elements.each { result.put(toCamelCase(it.alias), tuple.get(it.alias))}
        return result
    }


}
