package diego.db.collections

import com.google.inject.ImplementedBy
import diego.db.DTO;

/**
 * Interface for managing a collection of entities. A collection represents a table.
 *
 * @author Kevin Sheppard
 * Date: July 06, 2023
 */
interface EntityCollection<T, K extends Serializable> {
    /**
     * Find an entity by its ID.
     *
     * @param id The ID of the entity
     * @return An entity with the specified ID or null if no such entity exists.
     */
    T find(K id);

    /**
     * Delete an entity
     *
     * @param id The ID of the entity
     */
    void delete(K id);

    /**
     * Insert an entity. The ID will be written to the supplied object if it is marked with
     * <pre>@GeneratedValue</pre>
     *
     * @param obj An annotated JPA Entity
     */
    void insert(T instance);

    /**
     * Update a JPA entity
     *
     * @param instance An instance of T
     */
    void update(T instance);

    /**
     * Update a JPA entity using a data transfer object
     *
     * @param id The ID of the entity to update
     * @param dto The data transfer object
     */
    void update(K id, DTO dto);

    /**
     * Create a CollectionQuery using JPQL and named parameters
     *
     * @param jpql A JPQL query
     * @param namedParams Parameters for the query
     * @return A CollectionQuery<T>
     */
    CollectionQuery<T> createQuery(String jpql, Map<String, Object> namedParams);

    /**
     * Create a CollectionQuery using JPQL and positional parameters
     *
     * @param jpql A JPQL query
     * @param positionalParams Parameters for the query
     * @return A CollectionQuery<T>
     */
    CollectionQuery<T> createQuery(String jpql, Object... positionalParams);

    /**
     * @param jpql A JPQL query
     * @return A CollectionQuery<T>
     */
    CollectionQuery<T> createQuery(String jpql);

}
