package diego.init;

import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import com.google.inject.AbstractModule;
import com.google.inject.Module;
import com.typesafe.config.Config;
import com.typesafe.config.ConfigFactory;

import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

/**
 * Typed representation of application settings stored in *.conf files.
 */
public class Settings {
    private static Settings instance;
    private final Env env;
    private final HttpConfig http;
    private final Config config;
    private List<Module> modules;

    private Settings(Config config) {
        this.config = config;
        this.env = new Env();
        this.http = new HttpConfig();
    }

    /**
     * Non-thread safe singleton method.
     *
     * @return The Settings singleton.
     */
    public static Settings getInstance() {
        if(instance == null) load();
        return instance;
    }

    /**
     * Parse the configuration file and create an instance of Settings by loading the default configuration file
     * or the source specified in the config.file, config.resource or config.url System property.
     *
     * @return The loaded application settings.
     */
    public static Settings load() {
        instance = new Settings(ConfigFactory.load());
        return instance;
    }

    /**
     * Same as {@link Settings#load()}, but allows the user to specify a configuration file.
     *
     * @param configuration The name of the configuration resource e.g. "test.conf"
     * @return The loaded application settings.
     */
    public static Settings load(String configuration) {
        instance = new Settings(ConfigFactory.load(configuration));
        return instance;
    }

    public String getDateFormat() {
        return config.getString("app.dateFormat");
    }

    public HttpConfig getHttp() {
        return http;
    }

    public Env getEnv() {
        return env;
    }

    public List<Module> getModules() {
        if(modules != null) return modules;
        modules = new ArrayList<>();
        if(!config.hasPath("app.modules")) return modules;
        try {
            final List<String> moduleList = config.getStringList("app.modules");
            for(String className : moduleList) {
                Class<?> klass = Class.forName(className);
                Constructor<?> constructor = klass.getConstructor();
                Object obj = constructor.newInstance();
                if(obj instanceof AbstractModule) modules.add((AbstractModule) obj);
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return modules;
    }

    /**
     *
     * @return The root configuration object.
     */
    public Config getData() {
        return config;
    }

    /**
     *
     * @param path The root of a configuration tree. E.g. "app".
     * @return A Config object for the values at [path].
     */
    public Config path(String path) {
        return getData().getConfig(path);
    }
    /**
     * Utility class for reading the "app.env" config property.
     */
    public class Env {
        /**
         * @return True if it matches prod.*
         */
        public boolean isProd() {
            return environment().matches("prod.*");
        }

        /**
         * @return True if it matches dev.*
         */
        public boolean isDev() {
            return environment().matches("dev.*");
        }

        /**
         *
         * @param env The value to compare, e.g. staging.
         * @return True if it's a match.
         */
        public boolean is(String env) {
            return environment().equals(env);
        }

        /**
         * Shortcut for accessing the "mode" value from config.
         *
         * @return The value for app.env
         */
        private String environment() {
            return config.getString("app.env");
        }
    }

    class HttpConfig {
        private final Config config = path("app.http");

        public int getPort() {
            return config.getInt("port");
        }

        public long getMaxBodySize() {
            return config.getBytes("maxBodySize");
        }

        public String getFileUploadsDir() {
            return config.getString("fileUploadsDir");
        }

        public Boolean getAutoDeleteFiles() { return config.getBoolean("autoDeleteFiles"); }

    }

    //TODO: Move this into its own utility class
    static class JsonConfig {
        private final ObjectMapper objectMapper;

        public JsonConfig(ObjectMapper objectMapper) {
            this.objectMapper = objectMapper;
        }

        public JsonConfig disable(SerializationFeature feature) {
            objectMapper.disable(feature);
            return this;
        }

        public JsonConfig disable(DeserializationFeature feature) {
            objectMapper.disable(feature);
            return this;
        }

        /**
         * Manipulate the ObjectMapper directly if you wish. Go wild!
         *
         * @return The application's ObjectMapper instance.
         * @see <a href="https://fasterxml.github.io/jackson-databind/javadoc/2.7/com/fasterxml/jackson/databind/ObjectMapper.html">Read the docs</a>
         */
        public ObjectMapper getObjectMapper() {
            return objectMapper;
        }
    }
}
