package diego.templates

import com.google.inject.Inject
import com.google.inject.Provider
import diego.init.Settings
import diego.utils.FastRuntimeException
import freemarker.cache.ClassTemplateLoader
import freemarker.cache.FileTemplateLoader
import freemarker.cache.TemplateLoader
import freemarker.template.Configuration
import freemarker.template.TemplateExceptionHandler

/**
 * @author Kevin Sheppard
 * Date: Jul 05, 2023
 */
class FreemarkerConfigurationProvider implements Provider<Configuration> {
    private Configuration configuration

    @Inject
    void initialize(Settings settings, JsonMethod jsonMethod) {
        configuration = buildBaseConfiguration()
        configuration.setSharedVariable("json", jsonMethod)
        configuration.setTemplateLoader(createTemplateLoader(settings.data.getString("app.templates.dir")))
        configuration.templateExceptionHandler = settings.env.dev ?
                TemplateExceptionHandler.HTML_DEBUG_HANDLER :
                TemplateExceptionHandler.RETHROW_HANDLER
    }

    private static TemplateLoader createTemplateLoader(String uri) {
        String[] components = uri.split(':')
        if(components.length != 2) throw new FastRuntimeException('You must specify a scheme of either file: or classpath:')
        String scheme = components[0]
        String path = components[1].trim()
        path = path.endsWith('/') ? path : path + '/'
        switch (scheme) {
            case 'file':
                return new FileTemplateLoader(new File(path))
            case 'classpath':
                return new ClassTemplateLoader(FreemarkerConfigurationProvider, path)
            default:
                throw new FastRuntimeException('No scheme specified for template path: ' + uri)
        }

    }

    @Override
    Configuration get() {
        return configuration
    }

    static Configuration buildBaseConfiguration() {
        final Configuration configuration = new Configuration(Configuration.VERSION_2_3_31)
        configuration.tap {
            defaultEncoding = "UTF-8"
            logTemplateExceptions = false
            wrapUncheckedExceptions = true
            fallbackOnNullLoopVariable = false
        }
        return configuration
    }
}
