package diego.tooling


import diego.Application
import diego.Components
import diego.Diego
import diego.aop.UsesDependencyService
import groovy.util.logging.Slf4j
import org.gradle.tooling.BuildLauncher
import org.gradle.tooling.GradleConnectionException
import org.gradle.tooling.GradleConnector
import org.gradle.tooling.ProjectConnection

/**
 * @author Kevin Sheppard
 * Date: Mar 01, 2023
 */
@Slf4j
class CLILauncher extends Diego implements Runnable, UsesDependencyService {
  private final File projectDir
  private final String mainClassName
  private ProjectConnection projectConnection
  private Boolean serverStarted = false
  private Application application

  private CLILauncher(File projectDir, String mainClassName) {
    this.projectDir = projectDir
    this.mainClassName = mainClassName
  }

  /**
   *
   * @param args args[0]: The main class name
   */
  static void main(String[] args) {
    if (args.length == 0) {
      System.err.println("You must supply [main class name]")
      System.exit(1)
    }
    final File projectDirectory = new File(System.getProperty('user.dir'))
    final String mainClassName = args[0]
    if(!mainSourceFileExists(mainClassName)) {
      System.err.println("Could not find $mainClassName. Exiting.")
      System.exit(404)
    }
    final CLILauncher cliLauncher = new CLILauncher(projectDirectory, mainClassName)
    cliLauncher.launch()
  }

  private static Boolean mainSourceFileExists(String className) {
    final String classNameAsPath = className.replaceAll('\\.', '/')
    final File groovySource = new File("src/main/groovy/${classNameAsPath}.groovy")
    final File javaSource = new File("src/main/java/${classNameAsPath}.java")
    return (groovySource.exists() || javaSource.exists())
  }

  /**
   * Clears the router and applies the new routing configuration as per the loaded Application.
   */
  private void applyChanges() {
    router.clear()
    configureRouter(application)
    if(!serverStarted) {
      serverStarted = true
      start({
        log.info('Watching for code changes. Press Ctrl+C to exit.')
        application.onStart()
      })
    } else {
      log.info('Done.')
      application.onStart() // Running it here, since we can't run it as a callback on server start.
    }
  }

  @Override
  void run() {
    try {
      if(serverStarted) log.info('Updating application...')
      /* Trap any runtime issues and prevent starting the server or reloading the application */
      final FileSystemClassLoader fileSystemClassLoader = new FileSystemClassLoader(CLILauncher.classLoader, projectDir.path)
      Class<?> clazz = fileSystemClassLoader.loadClass(mainClassName)
      application = inject(clazz) as Application
      applyChanges()
    } catch (Exception e) {
      e.printStackTrace()
    }
  }

  private void launch() {
    bootstrap()
    try {
      final ReactiveOutputStream outputStream = new ReactiveOutputStream(this)
      log.info('Starting project build...')
      projectConnection = GradleConnector
          .newConnector()
          .forProjectDirectory(projectDir)
          .connect()
      final BuildLauncher launcher = projectConnection.newBuild()
          .forTasks('compileGroovy')
          .addArguments('--continuous')
          .setStandardOutput(outputStream)
          .setStandardError(System.err)

      /* Starting in a new thread so that we can launch the server in the main thread. */
      final Thread builderThread = new Thread({
        launcher.run()
      })
      builderThread.name = 'Gradle-Build'
      builderThread.start()
    } catch (GradleConnectionException | IllegalStateException e) {
      log.error(e.message, e.cause)
    } catch (InterruptedException ignored) {}
  }
}
