package diego.tooling

import groovy.util.logging.Slf4j

/**
 * A ClassLoader that loads classes from the file system. This classloader is what facilitates the hot reload
 * functionality.
 *
 * Author: Kevin Sheppard
 * Date: February 25, 2023
 */
@Slf4j
class FileSystemClassLoader extends ClassLoader {
  private final String projectDir
  private final Map<String, Class<?>> cache = new LinkedHashMap<>()

  FileSystemClassLoader(ClassLoader parent, String projectDir) {
    super(parent)
    this.projectDir = projectDir
  }

  @Override
  Class<?> loadClass(String name) throws ClassNotFoundException {
    log.trace("Loading {}", name)
    if(cache.containsKey(name)) return cache.get(name)
    /**
     * This is explained here:
     * https://jenkov.com/tutorials/java-reflection/dynamic-class-loading-reloading.html#:~:text=Designing%20your%20Code%20for%20Class%20Reloading
     *
     * Basically, we search our project's build output directory for the desired class, if it's
     * not there, the we kick it upstairs to the parent Classloader for loading.
     */
    try {
      final Optional<File> classFileOptional = findClassFile(name)
      if(classFileOptional.empty) {
        return super.loadClass(name)
      }
      final byte[] bytes = classFileOptional.get().bytes
      Class<?> definedClass = defineClass(name, bytes, 0, bytes.length)
      cache.put(name, definedClass)
      return definedClass
    }
    /**
     * Note: It is very important that ClassNotFound exceptions are thrown by this method.
     * This allows the class finding delegation logic to kick in. This is why we must be
     * very specific about the types of exceptions we catch here.
     */
    catch (ClassFormatError e) {
      e.printStackTrace()
      return null
    }
  }

  /**
   * Looks for a .class file in the project's build output directory.
   *
   * @param classFileName The file name in Java notation - com.example.Application
   * @return An Optional<File> which will be empty if the file doesn't exist.
   */
  private Optional<File> findClassFile(String classFileName) {
    String[] classFileNameComponents = classFileName.split('\\.')
    return ["groovy", "java"]
        .stream()
        .map {language ->
          final List<String> pathSegments = ['build', 'classes', language, 'main']
          pathSegments.addAll(classFileNameComponents)
          String filePath = resolve(pathSegments.toArray() as String[]) + '.class'
          new File(filePath)
        }
        .filter { file -> file.exists() }
        .findFirst()
  }

  /**
   *
   * @param pathSegments Segments of the file path e.g. foo/bar.txt => 'foo', 'bar.txt'
   * @return A file path relative to the project directory.
   */
  private String resolve(String... pathSegments) {
    final String separator = System.getProperty('file.separator')
    StringJoiner joiner = new StringJoiner(separator)
    joiner.add(projectDir)
    for(String segment : pathSegments) {
      joiner.add(segment)
    }
    return joiner.toString()
  }

}
