package diego.utils


import diego.Components
import diego.init.Settings
import diego.db.SessionFactoryProvider

/**
 * Utility class for managing the dependency container during testing.
 *
 * Unlike using the raw {@link Components} class, this one has caching logic that can be useful when building the
 * dependency container from a {@code @Before} or {@code @BeforeClass} context.
 *
 * <strong>To developers</strong> <br />
 * Beware of over-optimizing. It is better to rebuild the entire dependency container than, say, rebuild the Settings
 * class only. Singletons in other modules rely on the instance of Settings that existed at the point in time they
 * were created, and will not see any changes to the Settings singleton going forward.
 *
 * The SessionFactory, is one such class. Look at how the {@link SessionFactoryProvider} is implemented.
 *
 * @author Kevin Sheppard
 * Date: Mar 31, 2022
 */
class Helpers {
  private static String currentConfiguration

  /**
   * Initialize the dependency container. Will abort if the container has been already initialized.
   */
  static void initializeContext() {
    initializeContext("default")
  }

  /**
   * Initialize the dependency container but specify a configuration file to use. Will abort if the container has
   * already been initialized using the configuration file specified.
   *
   * @param configuration The configuration file to use, e.g. "test.conf".
   */
  static void initializeContext(String configuration) {
    if(currentConfiguration == configuration) return
    currentConfiguration = configuration
    Components.clear()
    loadSettings(configuration)
    Components.initializeFramework()
  }

  /**
   * Initialize Settings.
   *
   * @param configFile A configuration file e.g. "dev.conf" or "default" for the default configuration.
   */
  static void loadSettings(String configFile) {
    configFile == "default" ? Settings.load() : Settings.load(configFile)
  }

  /**
   * Retrieves an instance of T from the dependency container.
   * This method calls {@link Helpers#initializeContext()} if the container has not been initialized yet.
   *
   * @param klass The instance type.
   * @return An instance of T from the container.
   */
  static <T> T get(Class<T> klass) {
    if(!currentConfiguration) initializeContext()
    return Components.createInstance(klass)
  }

}
