package diego.utils

import java.beans.Introspector
import java.beans.PropertyDescriptor
import java.util.stream.Collectors

/**
 * @author Kevin Sheppard
 * Date: February 7, 2022 3:59AM
 */
class PropertyMapper {

    /**
     * Copies the values of one object to another, ignoring nulls.
     *
     * @param source The property source.
     * @param target The target object.
     */
    static void copyProperties(Object source, Object target) {
        var sourceProps = getPropertyDescriptors(source)
        var targetProps = getPropertyDescriptors(target)
        sourceProps.each { pd ->
            if(!pd.readMethod) return
            var value = pd.readMethod.invoke(source)
            if(value == null) return // Ignore nulls.
            var targetProp = findPropertyDescriptor(targetProps, pd.name)
            if(targetProp.present && targetProp.get().writeMethod) {
                var prop = targetProp.get()
                var method = prop.writeMethod
                method.invoke(target, value)
            }
        }
    }

    static Optional<PropertyDescriptor> findPropertyDescriptor(List<PropertyDescriptor> descriptors, String name) {
        return descriptors.stream().filter(pd -> pd.name == name).findFirst()
    }

    /**
     *
     * @param instance A Bean.
     * @return All usable PropertyDescriptors for instance.
     */
    static List<PropertyDescriptor> getPropertyDescriptors(Object instance) {
        var klass = instance.getClass()
        var descriptors = Introspector.getBeanInfo(klass).propertyDescriptors
        return Arrays.asList(descriptors).stream()
                .filter({pd -> pd.name != "class" })
                .filter({pd ->
                   try {
                       var field = klass.getDeclaredField(pd.name)
                       return field && !field.synthetic
                   } catch(NoSuchFieldException ignored) {
                       return false
                   }
                })
                .collect(Collectors.toList())
    }
}
