package diego.utils.httpclient

import groovy.util.logging.Slf4j
import org.apache.http.client.methods.CloseableHttpResponse
import org.apache.http.client.protocol.HttpClientContext
import org.apache.http.impl.client.CloseableHttpClient
import org.apache.http.impl.client.HttpClientBuilder
import org.apache.http.impl.client.LaxRedirectStrategy
import org.apache.http.protocol.BasicHttpContext
/**
 * A simple HTTP client for performing server to server communication.
 *
 * @author Kevin Sheppard
 * Date: Jan 05, 2022
 */
@Slf4j
class HttpClient {
  private String baseUrl

  /**
   * @param baseUrl A URL such as https://google.com
   * @return A new instance of HttpClient.
   */
  static HttpClient create(String baseUrl) {
    return new HttpClient(baseUrl: baseUrl)
  }

  /**
   * @return A CloseableHttpClient with sensible defaults configured.
   */
  private static CloseableHttpClient buildClient() {
    return HttpClientBuilder.create().setRedirectStrategy(new LaxRedirectStrategy()).build()
  }

  /**
   * @param path The path on the server e.g. /home.
   * @param method The HTTP Method.
   * @return A new {@link HttpClientRequest}.
   */
  HttpClientRequest createRequest(HttpMethod method, String path) {
    return new HttpClientRequest(baseUrl, path, method)
  }

  /**
   * @param method The HTTP method.
   * @return A new {@link HttpClientRequest} with the path set to "/".
   */
  HttpClientRequest createRequest(HttpMethod method) {
    return createRequest(method, "/")
  }

  /**
   *
   * @param path The path on the server e.g. /home
   * @return A new HTTP GET {@link HttpClientRequest}.
   */
  HttpClientRequest createRequest(String path) {
    return createRequest(HttpMethod.GET, path)
  }

  /**
   * @return A new HTTP GET request to the server root.
   */
  HttpClientRequest clientRequest() {
    return createRequest(HttpMethod.GET, "/")
  }

  /**
   * Send the request to the server.
   *
   * @param request An {@link HttpClientRequest}.
   * @return An {@link HttpClientResponse} or NULL if there was an error.
   */
  @SuppressWarnings('GrMethodMayBeStatic')
  HttpClientResponse send(HttpClientRequest request) {
    try (CloseableHttpClient client = buildClient()) {
      def httpUriRequest = request.toHttpUriRequest()
      def httpContext = new BasicHttpContext()
      httpContext.setAttribute(HttpClientContext.COOKIE_STORE, request.cookieStore)
      try(CloseableHttpResponse httpResponse = client.execute(httpUriRequest, httpContext)) {
        return HttpClientResponse.create(httpResponse)
      }
    } catch (IOException e) {
      log.error(e.message)
      return null
    }
  }

}
