package diego.utils.httpclient

import org.apache.http.HttpEntity
import org.apache.http.HttpHeaders
import org.apache.http.client.methods.HttpUriRequest
import org.apache.http.client.methods.RequestBuilder
import org.apache.http.client.utils.URIBuilder
import org.apache.http.entity.StringEntity
import org.apache.http.impl.client.BasicCookieStore
import org.apache.http.impl.cookie.BasicClientCookie
import diego.web.ContentTypes
/**
 * A class to store configuration details for an HTTP request.
 *
 * @author Kevin Sheppard
 * Date: Jan 05, 2022
 */
class HttpClientRequest {
  private HttpMethod method
  private final LinkedHashMap<String, String> headers = new LinkedHashMap<>()
  private URIBuilder uriBuilder
  protected final BasicCookieStore cookieStore = new BasicCookieStore()
  private Object body

  protected HttpClientRequest(String baseUrl, String path, HttpMethod method) {
    this.uriBuilder = new URIBuilder(baseUrl).setPath(path)
    this.method = method
  }

  /**
   * Add a header to the request.
   *
   * @param name Header name.
   * @param value Header value.
   * @return This instance.
   */
  HttpClientRequest setHeader(String name, String value) {
    headers.put(name, value)
    return this
  }

  /**
   * Set the path for the request.
   *
   * @param path The path for the request.
   * @return This instance.
   */
  HttpClientRequest setPath(String path) {
    uriBuilder.setPath(path)
    return this
  }

  /**
   * Set the HTTP method for the request.
   *
   * @param method The HTTP method.
   * @return This instance.
   */
  HttpClientRequest setMethod(HttpMethod method) {
    this.method = method
    return this
  }

  /**
   * Set a URL parameter.
   *
   * @param name The name of the parameter.
   * @param values One or more values for the parameter (as in the case of a checkbox array).
   * @return This instance.
   */
  HttpClientRequest setParam(String name, String... values) {
    values.each { value -> uriBuilder.addParameter(name, value) }
    return this
  }

  /**
   * Set the request body.
   *
   * @param body The body of the request. It should be an instance of {@link FormData} or String.
   * @return This instance.
   */
  HttpClientRequest setBody(Object body) {
    this.body = body
    return this
  }

  /**
   * Set a String as the request body and specify its content type using a String.
   * Common content types can be found in {@link ContentTypes}.
   *
   * @param body The request body.
   * @param contentType The Content-Type of the request.
   * @return This instance.
   */
  HttpClientRequest setBody(String body, String contentType) {
    this.body = body
    return setHeader(HttpHeaders.CONTENT_TYPE, contentType)
  }

  protected Object getBody() {
    return this.body
  }

  /**
   * Set a request cookie.
   *
   * @param name The name of the cookie.
   * @param value The cookie value.
   * @return This instance.
   */
  HttpClientRequest setCookie(String name, String value) {
    def cookie = new BasicClientCookie(name, value)
    cookie.setDomain(uriBuilder.host)
    cookie.setPath("/")
    cookieStore.addCookie(cookie)
    return this
  }

  /**
   * Converts this request into an Apache HttpUriRequest for use by the {@link HttpClient}.
   *
   * @return An Apache HttpUriRequest.
   */
  protected HttpUriRequest toHttpUriRequest() {
    RequestBuilder builder = RequestBuilder.create(method.methodName).setUri(uriBuilder.build())
    headers.each { header -> builder.addHeader(header.key, header.value) }
    if(body != null) builder.setEntity(entity)
    return builder.build()
  }

  /**
   * Logic for setting the right request body and headers.
   *
   * @param requestBase
   * @return The requestBase with the Entity enclosed.
   */
  private HttpEntity getEntity() {
    if(body instanceof FormData) {
      setHeader(HttpHeaders.CONTENT_TYPE, body.encType.mimeType)
      return body.entity
    }
    if(body instanceof String) {
      return new StringEntity(body)
    }
    throw new Exception("Unsupported body type: ${body.class}. Body should be one of String, MultipartForm or UrlEncodedForm")
  }
}
