package diego.utils.httpclient

import org.apache.http.HttpEntity
import org.apache.http.entity.ContentType
import org.apache.http.entity.mime.MultipartEntityBuilder

/**
 * Create a Multipart form used for uploading files.
 *
 * @author Kevin Sheppard
 * Date: Jan 06, 2022
 */
class MultiPartForm extends FormData {

  // If this stops working, use entityBuilder.setMode(HttpMultipartMode.BROWSER_COMPATIBLE)
  private final MultipartEntityBuilder entityBuilder = MultipartEntityBuilder.create()

  /**
   * Add a parameter to the form.
   *
   * @param name The name of the form field.
   * @param values One or more values (for checkbox arrays).
   * @return This instance.
   */
  @Override
  MultiPartForm addParameter(String name, String... values) {
    if(values.length == 0) return this
    Arrays.asList(values).stream()
        .filter({ it != null })
        .forEach({ value -> entityBuilder.addTextBody(name, value) })
    return this
  }

  /**
   * Add a file attachment.
   *
   * @param name The name of the field.
   * @param file The file.
   * @return This instance.
   */
  MultiPartForm addParameter(String name, File file) {
    if(file == null) return this
    entityBuilder.addBinaryBody(name, file)
    return this
  }

  /**
   * Convenience method for adding parameters to the form from a Map.
   *
   * @param params The parameters to add to the form.
   * @return This instance.
   */
  MultiPartForm withParameters(Map<String, Object> params) {
    params.entrySet().stream()
        .filter({ it.value != null })
        .forEach({entry ->
          if(entry.value instanceof String) addParameter(entry.key, (String) entry.value)
          else if(entry.value instanceof File) addParameter(entry.key, (File) entry.value)
          else throw new Exception("Unsupported entry type: ${entry.value.class}.")
        })
    return this
  }

  @Override
  protected ContentType getEncType() {
    return ContentType.MULTIPART_FORM_DATA
  }

  @Override
  protected HttpEntity getEntity() {
    return entityBuilder.build()
  }
}
