package diego.validation

import diego.aop.ApplicationDependencyContainerService
import diego.aop.DependencyContainerService
import diego.utils.FastRuntimeException
import diego.web.HttpServerRequest
import jakarta.validation.ConstraintViolation
import jakarta.validation.Validator

import java.util.stream.Collectors

/**
 * The data validation feature of the framework.
 *
 * @author Kevin Sheppard
 * Date: May 12, 2023
 */
class Form<T> {
    private Class<T> modelType
    private Validator validator
    protected static DependencyContainerService dependencyContainerService = new ApplicationDependencyContainerService()

    static <T> Form<T> create(Class<T> modelClass) {
        final Validator validator = dependencyContainerService.getBean(Validator)
        return new Form(modelClass, validator)
    }

    protected Form(Class<T> modelType, Validator validator) {
        this.modelType = modelType
        this.validator = validator
    }

    T parse(HttpServerRequest request) {
        final var contentType = request.contentType.toLowerCase()
        RequestParser requestParser = findRequestParser(contentType)
        T model = requestParser.parseRequest(request, modelType)
        Set<ConstraintViolation<T>> violations = validator.validate(model)
        if(!violations.empty) throw new FormValidationException(group(violations))
        return model
    }

    private static findRequestParser(String contentType) {
        if(contentType.contains('json')) return dependencyContainerService.getBean(JsonRequestParser)
        if(contentType.contains('form')) return new FormRequestParser()
        throw new FastRuntimeException('No request parser was found for Content-Type: ' + contentType)
    }

    protected static LinkedHashMap<String, List<String>> group(Set<ConstraintViolation> violations) {
        return violations.stream()
            .map({ violation -> [ propertyPath: violation.propertyPath, message: violation.message ] })
            .collect(Collectors.groupingBy({ map -> map['propertyPath'] } ,
                Collectors.mapping({ map -> map['message']}, Collectors.toList())))
    }

}
