package diego.validation

import diego.aop.UsesDependencyService

import static java.util.stream.Collectors.groupingBy
import static java.util.stream.Collectors.mapping

import diego.web.HttpServerRequest
import groovy.util.logging.Slf4j
import io.vertx.core.MultiMap
import io.vertx.ext.web.FileUpload
import org.minimalcode.beans.ObjectWrapper

import static java.util.stream.Collectors.toList

/**
 * Request parser for form requests.
 *
 * @author Kevin Sheppard
 * Date: May 12, 2023
 */
@Slf4j
class FormRequestParser implements RequestParser, UsesDependencyService {

  @Override
  <T> T parseRequest(HttpServerRequest request, Class<T> klass) {
    final T instance = inject(klass)
    copyProperties(request.body.formParams, instance)
    if(isMultipart(request)) handleUploads(request, instance)
    return instance
  }

  static void copyProperties(MultiMap params, Object target) {
    ObjectWrapper wrapper = new ObjectWrapper(target)
    var visitedKeys = []
    params.entries().forEach(entry -> {
      if (visitedKeys.contains(entry.key)) return
      if (entry.key.endsWith("[]")) {
        var value = params.getAll(entry.key)
        var propertyName = entry.key.replace("[]", "")
        wrapper.setValue(propertyName, value)
      } else {
        wrapper.setValue(entry.key, entry.value)
      }
      visitedKeys.add(entry.key)
    })
  }

  static void handleUploads(HttpServerRequest request, Object target) {
    /* The Set<FileUploads> needs to be grouped according to parameter name. Once we have this grouping, we can then
    * determine what needs to be handled as a collection and what needs to be handled as a single value. */
    final Map<String, List<FileUpload>> uploads = request.fileUploads.stream()
        .filter({ it.size() > 0l })
        .map(upload -> [param: upload.name(), upload: upload])
        .collect(groupingBy((Map tuple) -> tuple['param'], mapping((Map tuple) -> tuple['upload'], toList())))
    ObjectWrapper wrapper = new ObjectWrapper(target)
    uploads.forEach((key, val) -> {
      if(val.empty) return
      if(key.endsWith('[]')) {
        final String _key = key.replace('[]', '')
        wrapper.setValue(_key, val)
      } else wrapper.setValue(key, val[0])
    })
  }

  static boolean isMultipart(HttpServerRequest request) {
    final String contentType = request.headers.get('Content-Type')
    if(!contentType) return false
    return contentType.containsIgnoreCase('multipart')
  }
}
