package diego.web;

import io.vertx.core.MultiMap;
import io.vertx.ext.web.FileUpload;

import java.net.URI;
import java.util.Map;
import java.util.Set;

/**
 * @author Kevin Sheppard
 * Date: Nov 02, 2021
 */
public interface HttpServerRequest {

  /**
   * Shortcut for HttpServerRequest#getState().get(String)
   *
   * @param key The state item.
   * @param <T> The return type.
   * @return An item from the request state, or null if it is not found.
   */
  <T> T get(String key);

  /**
   * Facilitates the use of Groovy's subscript operator ([]).
   * @see HttpServerRequest#get(String)
   *
   * @param key The state item.
   * @param <T> The return type.
   * @return An item from the request state, or null if it is not found.
   */
  <T> T getAt(String key);

  /**
   *
   * @return The request body.
   */
  Body getBody();

  /**
   * @return The value of the Content-Type header.
   */
  String getContentType();

  /**
   * RFC States that there can be multiple cookies with the same name, but
   * seeing that this is a practice we try to discourage internally, this pattern
   * will go a long way in enforcing that.
   *
   * Name your cookies uniquely!
   *
   * The Map will contain the cookie name and its value.
   *
   * @return A Map of Cookies and their values.
   */
  Map<String, String> getCookies();

  /**
   * @return The files sent with this request.
   */
  Set<FileUpload> getFileUploads();

  MultiMap getHeaders();

  String getHost();

  /**
   *
   * @return The HTTP method used for this request.
   */
  String getMethod();

  /**
   *
   * @return A Map of path parameters.
   */
  Map<String, String>  getPathParams();

  /**
   * @return The query parameters in the URL.
   */
  MultiMap getQueryParams();

  /**
   *
   * @return The path portion of the URI e.g. /customers
   */
  String getPath();

  /**
   * @param name The name of the parameter.
   * @return The named parameter from the path e.g. /users/:id
   */
  String getPathParam(String name);

  Map<String, Object> getState();

  URI getUri();

  /**
   * Pass the request on to the next matching route. Useful for filters.
   */
  void next();

  /**
   * Add an arbitrary item to the request state.
   *
   * @param key The key.
   * @param value The value.
   * @return This instance.
   */
  HttpServerRequest put(String key, Object value);

  /**
   * Facilitates the use of Groovy's subscript operator ([]).
   *
   * @see HttpServerRequest#put(String, Object)
   *
   * @param key The key.
   * @param value The value.
   */
  void putAt(String key, Object value);

  /**
   * Remove an item from the request's state.
   *
   * @param key The item key.
   * @return This instance.
   */
  HttpServerRequest remove(String key);

  /**
   * If you find yourself using this method, document your use case and create a pull request.
   *
   * @return The underlying Vertx Web HttpServerRequest.
   * @see <a href="https://vertx.io/docs/apidocs/io/vertx/core/http/HttpServerRequest.html">HttpServerRequest</a>
   */
  io.vertx.core.http.HttpServerRequest unwrap();

  /**
   * Serves to encapsulate all logic relating to the request body.
   */
  interface Body {
    /**
     *
     * @return The request's body as a String.
     */
    String toString();

    /**
     *
     * @return The parameters from the submitted form. Can be a URLEncoded or Multipart form.
     */
    MultiMap getFormParams();

    /**
     *
     * @return The request body as a usable Map (only for JSON requests).
     */
    Map<String, Object> getData();
  }
}
