package diego.web;

import java.util.Map;

/**
 * Interface created to guide the development of the HttpResponse object.
 *
 * @author Kevin Sheppard
 * Date: Nov 02, 2021
 */
public interface HttpServerResponse {
  public static final int HTTP_OK = 200;
  public static final int HTTP_ERROR = 500;
  public static final int HTTP_BAD_REQUEST = 400;
  public static final int HTTP_NOT_FOUND = 404;
  public static final int HTTP_UNAUTHORIZED = 401;
  public static final int HTTP_CREATED = 201;
  public static final int HTTP_REDIRECT_SEE_OTHER = 303;

  HttpServerResponse removeCookie(String name);

  void render(String templateName);
  void render(String templateName, Map<String, Object> data);

  /**
   * Restarts the router with the new path, preserving all headers, cookies, state and parameters.
   * It's equivalent to servlet forwarding and <b>not</b> an HTTP redirect.
   *
   * @param path The path of the other handler.
   */
  void reroute(String path);

  /**
   * Set the Content-Type header of the response. Use one of
   * the constants in {@link ContentTypes} or a custom valid
   * content type.
   *
   * @param contentType The Content-Type value.
   * @return This instance.
   */
  HttpServerResponse setContentType(String contentType);

  /**
   * Create a cookie that lasts for a month.
   * Common durations can be found as constants in the {@link Cookie} class.
   *
   * @param name The name of the cookie.
   * @param value The cookie's value.
   * @return This instance.
   */
  HttpServerResponse setCookie(String name, String value);

  /**
   * Create a cookie.
   *
   * @param cookie An instance of {@link Cookie}. Create it by calling {@link Cookie#create(String, String)}.
   * @return This instance.
   */
  HttpServerResponse setCookie(Cookie cookie);

  HttpServerResponse setHeader(String name, String value);

  /**
   * Set the status code of the response. Common status codes
   * can be found as HTTP_* constants of {@link HttpServerResponse}.
   *
   * @param statusCode The HTTP status code.
   * @return This instance.
   */
  HttpServerResponse setStatusCode(int statusCode);
  HttpServerResponse setStatusMessage(String statusMessage);

  /**
   * Send String content back to the client.
   *
   * @param content The content to send.
   * @param contentType A valid HTTP Content Type. See {@link ContentTypes} for common values.
   */
  void send(String content, String contentType);
  void send(String content);

  /**
   * Send a response back to the client.
   */
  void send();

  /**
   * Send a file back to the client.
   *
   * @param fileName The name of the file. Can be a classpath resource or an actual file on disk.
   */
  void sendFile(String fileName);

  /**
   * Send a file back to the client.
   *
   * @param fileName The name of the file. Can be a classpath resource or an actual file on disk.
   * @param contentType A valid HTTP Content Type. See {@link ContentTypes} for common values.
   */
  void sendFile(String fileName, String contentType);

  /**
   * Send a JSON response to the client.
   *
   * @param data The object to serialize.
   */
  void sendJson(Object data);

  /**
   * Use with caution as this method will be removed in future updates. If you find yourself using
   * this method, make a note of your use case and submit a pull request to the main repo.
   *
   * @return The underlying HttpServerResponse from the Vertx Web Framework.
   * @see <a href="https://vertx.io/docs/apidocs/io/vertx/core/http/HttpServerResponse.html">HttpServerResponse</a>
   */
  io.vertx.core.http.HttpServerResponse unwrap();

  HttpServerResponse badRequest();
  HttpServerResponse created();
  HttpServerResponse error();
  HttpServerResponse notFound();
  HttpServerResponse ok();

  /**
   * Perform a 303 redirect to another URL.
   *
   * @param url The URL to redirect to.
   */
  void redirect(String url);
  HttpServerResponse todo();
  HttpServerResponse unauthorized();

}
