package diego.web

import groovy.lang.Closure;
import groovy.lang.DelegatesTo
import io.vertx.ext.web.RoutingContext;
import io.vertx.ext.web.handler.StaticHandler

import java.util.function.Consumer
import java.util.regex.Pattern;

/**
 * Interface created to guide the development of the Router.
 *
 * @author Kevin Sheppard
 * Date: Nov 02, 2021
 */
public interface Router {

 /*
  * Ideally, we would like it to be get(String path, Handler... handlers) but Groovy does not understand how to
  * deconstruct varargs parameters when they are passed using the closure/lambda syntax.
  * See this thread on StackOverflow and this Jira issue for details:
  * https://stackoverflow.com/questions/41311593/how-do-i-pass-interfaces-as-varargs-parameters-to-a-method-in-groovy
  * https://issues.apache.org/jira/browse/GROOVY-10529
  *
  * A very unfortunate flaw of the language that I hope can be fixed in an upcoming release. If not, we'll take a look
  * at Kotlin, which doesn't have this issue, and would allow us to tighten up the Router API a bit.
  */

 /**
  * Create a GET route.
  *
  * @param path The route path.
  * @return a GET route.
  */
 Route get(String path)

 /**
  * Create a GET route and set a handler.
  *
  * @param path The route path.
  * @param handler A handler.
  * @return A GET route.
  */
 Route get(String path, Handler handler)

 /**
  * Create a DELETE route.
  *
  * @param path The route path.
  * @return a DELETE route.
  */
 Route delete(String path)

 /**
  * Create a DELETE route and set a handler.
  *
  * @param path The route path.
  * @param handler A handler.
  * @return A DELETE route.
  */
 Route delete(String path, Handler handler)

 /**
  * Create a PATCH route.
  *
  * @param path The route path.
  * @return a PATCH route.
  */
 Route patch(String path)

 /**
  * Create a PATCH route and set a handler.
  *
  * @param path The route path.
  * @param handler A handler.
  * @return A PATCH route.
  */
 Route patch(String path, Handler handler)

 /**
  * Create a POST route.
  *
  * @param path The route path.
  * @return a POST route.
  */
 Route post(String path)

 /**
  * Create a POST route and set a handler.
  *
  * @param path The route path.
  * @param handler A handler.
  * @return A POST route.
  */
 Route post(String path, Handler handler)

 /**
  * Create a PUT route.
  *
  * @param path The route path.
  * @return a PUT route.
  */
 Route put(String path)

 /**
  * Create a PUT route and set a handler.
  *
  * @param path The route path.
  * @param handler A handler.
  * @return A PUT route.
  */
 Route put(String path, Handler handler)

 /**
  * Creates an interceptor for a specific path.
  *
  * @param path The path to intercept. E.g. /private/*
  * @param handler A handler.
  */
 void before(String path, Handler handler)

 /**
  * Creates an interceptor that intercepts all routes following this one.
  *
  * @param handler A handler.
  */
 void before(Handler handler)

 /**
  * Creates an interceptor using regex as a path pattern.
  *
  * @param path The path pattern. E.g. ^/private-.*
  * @param handler A handler.
  */
 void beforeRegex(String path, Handler handler)

 /**
  * Serve a directory using a static files handler.
  *
  * @param path A path.
  * @param handlerOptions Configuration for the static files handler.
  * @return A GET route.
  */
 void path(String path, StaticHandlerOptions handlerOptions)

 /**
  * Use a RegEx pattern to specify a path at which to mount the static files handler.
  *
  * Uses a String as the pattern to match the Vert.x router method signature.
  *
  * @param regex A RegEx pattern.
  * @param handler A handler.
  * @return A GET route.
  */
 void pathRegex(String regex, StaticHandlerOptions handlerOptions)

 /**
  * Mount a sub-router at a particular path.
  *
  * @param prefix A path prefix such as /api/*
  * @param consumerClass A Class that implements java.util.function.Consumer<Router>
  */
 void path(String prefix, Class<? extends Consumer<Router>> consumerClass)

 /**
  * Mount a Vert.x Web handler to the routing stack at a specified path.
  *
  * @param path The path/path prefix
  * @param handler A Vert.x Web handler like CORSHandler
  */
 void path(String path, io.vertx.core.Handler<RoutingContext> handler)

 /**
  * Mount a Vert.x Web handler to the routing stack.
  *
  * @param handler A Vert.x Web handler like CORSHandler
  */
 void use(io.vertx.core.Handler<RoutingContext> handler)
}
