package diego.web

import diego.init.Settings
import groovy.transform.ToString
import io.vertx.ext.web.handler.StaticHandler

import java.time.Duration
import java.time.temporal.ChronoUnit
import java.util.function.Consumer
import java.util.regex.Matcher
import java.util.regex.Pattern

/**
 * Author: Kevin Sheppard
 * Date: April 01, 2023
 */
@ToString(excludes = ['delegate'])
class StaticHandlerOptions implements Consumer<StaticHandler> {
    /**
     * The directory to serve (relative to CWD)
     */
    String dir
    private Boolean cachingEnabled = true
    private long maxAge = parseDurationExpression('1d')
    private Boolean enableDirectoryListing = false
    private int maxCacheSize = 20

    static StaticHandlerOptions create(String dir, Settings.Env env) {
        final StaticHandlerOptions options = new StaticHandlerOptions(dir: dir)
        /* If we're in dev mode, let's disable caching */
        if (env.dev) {
            options.tap {
                cachingEnabled = false
                maxCacheSize = 0
            }
        }
        return options
    }

    @Override
    void accept(StaticHandler staticHandler) {
        staticHandler
                .setCachingEnabled(cachingEnabled)
                .setDirectoryListing(enableDirectoryListing)
        if (cachingEnabled)
            staticHandler
                    .setMaxCacheSize(maxCacheSize)
                    .setMaxAgeSeconds(maxAge)
                    .setCacheEntryTimeout(maxAge)
    }

    StaticHandler getDelegate() {
        return delegate
    }

    void setEnableDirectoryListing(Boolean enableDirectoryListing) {
        this.enableDirectoryListing = enableDirectoryListing
    }

    void setCachingEnabled(Boolean cachingEnabled) {
        this.cachingEnabled = cachingEnabled
    }

    void setMaxCacheSize(int maxCacheSize) {
        this.maxCacheSize = maxCacheSize
    }

    /**
     * Set the Max-Age header value using a duration expression in the form
     *
     * <amount><timeunit> where amount is any number and timeunit is one of:
     *
     * s = seconds
     * m = minutes
     * h = hours
     * d = days
     *
     * e.g. 1h for 1 hour
     *
     * Max: 365 days (365d)
     *
     * @param durationExpression A duration expression.
     */
    void setMaxAge(String durationExpression) {
        maxAge = parseDurationExpression(durationExpression)
    }

    /**
     * Parses a duration expression and returns a long value representing the duration in seconds.
     *
     * @param expression A duration expression.
     * @return a long value representing the duration in seconds.
     */
    static long parseDurationExpression(String expression) {
        final Pattern pattern = Pattern.compile('^([\\d]{1,})([smhd])$')
        final Matcher matcher = pattern.matcher(expression)
        if (!(matcher.matches())) throw new UnknownFormatConversionException("$expression is not a valid duration expression.")
        final long amount = Long.parseLong(matcher.group(1))
        final String sUnit = matcher.group(2)
        ChronoUnit unit = null
        switch (sUnit) {
            case 's':
                unit = ChronoUnit.SECONDS
                break;
            case 'm':
                unit = ChronoUnit.MINUTES
                break;
            case 'h':
                unit = ChronoUnit.HOURS
                break;
            case 'd':
                if (amount > 365) throw new IllegalArgumentException('Cache headers cannot be set longer than one year.')
                unit = ChronoUnit.DAYS
                break;
        }
        if (!unit) throw new UnknownFormatConversionException('Unable to determine duration from \'' + expression + '\'.')
        return Duration.of(amount, unit).toSeconds()
    }

}
