package diego.web.impl

import com.fasterxml.jackson.core.JsonProcessingException
import com.fasterxml.jackson.databind.ObjectMapper
import com.google.inject.Inject
import diego.web.HttpServerRequest
import groovy.util.logging.Slf4j
import io.vertx.core.MultiMap
import io.vertx.ext.web.FileUpload
import io.vertx.ext.web.RoutingContext
import io.vertx.core.http.HttpServerRequest as VRequest

/**
 * @author Kevin Sheppard
 * Date: Nov 02, 2021
 */
@Slf4j
class HttpServerRequestImpl implements HttpServerRequest {
  private RoutingContext ctx
  private VRequest request
  private ObjectMapper objectMapper
  private Body body

  @Inject
  HttpServerRequestImpl(ObjectMapper objectMapper) {
    this.objectMapper = objectMapper
  }

  @Override
  public <T> T get(String key) {
    return ctx.get(key)
  }

  @Override
  public <T> T getAt(String key) {
    return get(key)
  }

  @Override
  Body getBody() {
    return this.body
  }

  @Override
  String getContentType() {
    return request.getHeader("Content-Type")
  }

  @Override
  Map<String, String> getCookies() {
    return request.cookies().collectEntries(cookie -> [(cookie.name): cookie.value ])
  }

  @Override
  Set<FileUpload> getFileUploads() {
    return ctx.fileUploads()
  }

  @Override
  MultiMap getHeaders() {
    return request.headers()
  }

  @Override
  String getHost() {
    return request.host()
  }

  @Override
  String getMethod() {
    return request.method()
  }

  @Override
  Map<String, String> getPathParams() {
    return ctx.pathParams()
  }

  @Override
  MultiMap getQueryParams() {
    return request.params()
  }

  @Override
  String getPath() {
    return request.path()
  }

  @Override
  String getPathParam(String name) {
    return ctx.pathParam(name)
  }

  @Override
  Map<String, Object> getState() {
    return ctx.data()
  }

  @Override
  URI getUri() {
    return new URI(request.absoluteURI())
  }

  @Override
  void next() {
    ctx.next()
  }

  @Override
  HttpServerRequest put(String key, Object value) {
    state.put(key, value)
    return this
  }

  @Override
  void putAt(String key, Object value) {
    put(key, value)
  }

  @Override
  HttpServerRequest remove(String key) {
    state.remove(key)
    return this
  }

  HttpServerRequestImpl withRouteContext(RoutingContext routingContext) {
    this.ctx = routingContext
    this.request = routingContext.request()
    if(this.method.toUpperCase() in ["POST", "PATCH", "PUT"]) this.body = new RequestBodyImpl()
    return this
  }

  @Override
  io.vertx.core.http.HttpServerRequest unwrap() {
    return request
  }

  private class RequestBodyImpl implements Body {

    @Override
    String toString() {
      return ctx.body().asString()
    }

    @Override
    MultiMap getFormParams() {
      return request.formAttributes()
    }

    @Override
    Map<String, Object> getData() {
      try {
        objectMapper.readValue(toString(), LinkedHashMap)
      } catch (JsonProcessingException e) {
        log.error(e.message)
        return [:]
      }
    }
  }
}
