package diego.web.impl


import com.fasterxml.jackson.databind.ObjectMapper
import com.google.inject.Inject
import diego.templates.RequestModel
import diego.templates.TemplateEngine
import diego.web.ContentTypes
import diego.web.Cookie
import diego.web.HttpServerResponse
import groovy.util.logging.Slf4j
import io.vertx.core.http.Cookie as VCookie
import io.vertx.core.http.HttpServerResponse as VertxHttpServerResponse
import io.vertx.ext.web.RoutingContext

/**
 * @author Kevin Sheppard
 * Date: Nov 02, 2021
 */
@Slf4j
class HttpServerResponseImpl implements HttpServerResponse {
  private RoutingContext ctx
  private VertxHttpServerResponse response
  private final ObjectMapper objectMapper
  private final TemplateEngine templateEngine

  @Inject
  HttpServerResponseImpl(ObjectMapper objectMapper, TemplateEngine templateEngine) {
    this.objectMapper = objectMapper
    this.templateEngine = templateEngine
  }

  @Override
  HttpServerResponse removeCookie(String name) {
    response.removeCookie(name)
    return this
  }

  @Override
  void render(String templateName) {
    render(templateName, [:])
  }

  @Override
  void render(String templateName, Map<String, Object> data) {
    data.putAll(ctx.data())
    data.put("request", RequestModel.create(ctx.request()))
    String result = templateEngine.render("views/$templateName", data)
    send(result)
  }

  @Override
  void reroute(String path) {
    ctx.reroute(path)
  }

  @Override
  HttpServerResponse setContentType(String contentType) {
    response.putHeader('Content-Type', contentType)
    return this
  }

  @Override
  HttpServerResponse setCookie(String name, String value) {
    response.addCookie(VCookie.cookie(name, value))
    return this
  }

  @Override
  HttpServerResponse setCookie(Cookie cookie) {
    def vCookie = VCookie.cookie(cookie.name, cookie.value).setMaxAge(cookie.maxAge)
    response.addCookie(vCookie)
    return this
  }

  @Override
  HttpServerResponse setHeader(String name, String value) {
    response.putHeader(name, value)
    return this
  }

  @Override
  HttpServerResponse setStatusCode(int statusCode) {
    response.statusCode = statusCode
    return this
  }

  @Override
  HttpServerResponse setStatusMessage(String statusMessage) {
    response.setStatusMessage(statusMessage)
    return this
  }

  @Override
  void send(String content, String contentType) {
    setContentType(contentType)
    response.send(content)
  }

  @Override
  void send(String content) {
    response.send(content)
  }

  @Override
  void send() {
    response.send()
  }

  @Override
  void sendFile(String fileName) {
    response.sendFile(fileName)
  }

  @Override
  void sendFile(String fileName, String contentType) {
    this.setContentType(contentType).sendFile(fileName)
  }

  @Override
  void sendJson(Object obj) {
    this.setContentType(ContentTypes.JSON)
    if(obj instanceof String) { // If it's a JSON string, then set the content type and send it.
      send((String) obj)
      return
    }
    // We can afford not to wrap this in a try-catch because the framework handles displaying and logging errors for us.
    send(objectMapper.writeValueAsString(obj))
  }

  @Override
  io.vertx.core.http.HttpServerResponse unwrap() {
    return response
  }

  /* Convenience methods for setting the status code. */

  @Override
  HttpServerResponse badRequest() {
    return setStatusCode(HTTP_BAD_REQUEST)
  }

  @Override
  HttpServerResponse created() {
    return setStatusCode(HTTP_CREATED)
  }

  @Override
  HttpServerResponse error() {
    return setStatusCode(HTTP_ERROR)
  }

  @Override
  HttpServerResponse notFound() {
    return setStatusCode(HTTP_NOT_FOUND)
  }

  @Override
  HttpServerResponse ok() {
    return setStatusCode(HTTP_OK)
  }

  void redirect(String url) {
    setStatusCode(HTTP_REDIRECT_SEE_OTHER).setHeader("Location", url).send()
  }

  @Override
  HttpServerResponse todo() {
    return setStatusCode(501).setStatusMessage("TODO")
  }

  @Override
  HttpServerResponse unauthorized() {
    return setStatusCode(HTTP_UNAUTHORIZED)
  }

  HttpServerResponseImpl withRouteContext(RoutingContext routingContext) {
    this.ctx = routingContext
    this.response = routingContext.response()
    return this
  }
}
