package diego.web.impl


import diego.Components
import diego.aop.UsesDependencyService
import diego.validation.FormValidationException
import diego.web.Handler
import diego.web.HttpServerRequest
import diego.web.HttpServerResponse
import diego.web.Route
import groovy.util.logging.Slf4j
import io.vertx.ext.web.Route as VRoute
import io.vertx.ext.web.RoutingContext

import java.lang.reflect.UndeclaredThrowableException

/**
 * @author Kevin Sheppard
 * Date: Nov 02, 2021
 */
@Slf4j
class RouteImpl implements Route, UsesDependencyService {

  final VRoute delegate

  private RouteImpl(VRoute delegate) {
    this.delegate = delegate
  }

  static Route createRoute(VRoute vertxRoute) {
    return new RouteImpl(vertxRoute)
  }

  private static void sendErrorResponse(RoutingContext ctx, Throwable throwable) {
    ctx.response().setStatusCode(500).send("Unexpected error.")
    log.error(throwable.message, throwable)
  }

  @Override
  Route handler(Handler handler) {
    delegate.blockingHandler( { ctx ->
      final var request = createRequest(ctx)
      final var response = createResponse(ctx)
      try {
        handler.handle(request, response)
      }
      catch(FormValidationException e) {
        response.setStatusCode(400).sendJson([
            statusCode: 400,
            message: e.message,
            errors: e.constraintViolations
        ])
      }
      catch(Exception e) {
        final var throwable = e instanceof UndeclaredThrowableException ? e.cause : e
        sendErrorResponse(ctx, throwable)
      }
    }).failureHandler(ctx -> {
      sendErrorResponse(ctx, ctx.failure())
    })
    return this
  }

  /**
   * Create a HttpServerRequest using a RoutingContext delegate.
   *
   * @param ctx a <a href="https://vertx.io/docs/apidocs/io/vertx/ext/web/RoutingContext.html">RoutingContext</a> to serve as a delegate.
   * @return an instance of {@link HttpServerRequestImpl}
   */
  private HttpServerRequest createRequest(RoutingContext ctx) {
    return inject(HttpServerRequestImpl).withRouteContext(ctx)
  }


  private HttpServerResponse createResponse(RoutingContext ctx) {
    return inject(HttpServerResponseImpl).withRouteContext(ctx)
  }
}
