package diego.web.impl

import com.google.inject.Inject
import diego.aop.UsesDependencyService
import diego.web.Handler
import diego.web.Route
import diego.web.Router
import diego.web.StaticHandlerOptions
import io.vertx.ext.web.Router as VertxRouter
import io.vertx.ext.web.RoutingContext
import io.vertx.ext.web.handler.StaticHandler

import java.util.function.Consumer

import static RouteImpl.createRoute

/**
 * An implementation of the Router interface that leverages the Vertx Router.
 *
 * @author Kevin Sheppard
 * Date: Nov 02, 2021
 */
class RouterImpl implements Router, UsesDependencyService {
  private VertxRouter delegate

  RouterImpl() {}

  @Inject
  RouterImpl(io.vertx.ext.web.Router delegate) {
    this.delegate = delegate
  }

  VertxRouter getDelegate() {
    return delegate
  }

  @Override
  Route get(String path) {
    return createRoute(delegate.get(path))
  }

  @Override
  Route get(String path, Handler handler) {
    return get(path).handler(handler)
  }

  @Override
  Route delete(String path) {
    return createRoute(delegate.delete(path))
  }

  @Override
  Route delete(String path, Handler handler) {
    return delete(path).handler(handler)
  }

  @Override
  Route patch(String path) {
    return createRoute(delegate.patch(path))
  }

  @Override
  Route patch(String path, Handler handler) {
    return patch(path).handler(handler)
  }

  @Override
  Route post(String path) {
    return createRoute(delegate.post(path))
  }

  @Override
  Route post(String path, Handler handler) {
    return post(path).handler(handler)
  }

  Route put(String path) {
    return createRoute(delegate.put(path))
  }

  @Override
  Route put(String path, Handler handler) {
    return put(path).handler(handler)
  }

  @Override
  void before(String path, Handler handler) {
    createRoute(delegate.route(path)).handler(handler)
  }

  @Override
  void before(Handler handler) {
    createRoute(delegate.route()).handler(handler)
  }

  @Override
  void beforeRegex(String pattern, Handler handler) {
    createRoute(delegate.routeWithRegex(pattern)).handler(handler)
  }

  private static StaticHandler createStaticHandler(StaticHandlerOptions handlerOptions) {
    final StaticHandler staticHandler = StaticHandler.create(handlerOptions.dir)
    handlerOptions.accept(staticHandler)
    staticHandler
  }

  @Override
  void path(String path, StaticHandlerOptions handlerOptions) {
    delegate.get(path).handler(createStaticHandler(handlerOptions))
  }


  @Override
  void pathRegex(String regex, StaticHandlerOptions handlerOptions) {
    delegate.getWithRegex(regex).handler(createStaticHandler(handlerOptions))
  }

  @Override
  void path(String path, io.vertx.core.Handler<RoutingContext> handler) {
    delegate.route(path).handler(handler)
  }

  @Override
  void path(String prefix, Class<? extends Consumer<Router>> consumerClass) {
    final Consumer<Router> consumer = inject(consumerClass)
    final var childRouter = inject(RouterImpl)
    consumer.accept(childRouter)
    delegate.route(prefix).subRouter(childRouter.delegate)
  }

  @Override
  void use(io.vertx.core.Handler<RoutingContext> handler) {
    delegate.route().handler(handler)
  }
}
